#!/usr/bin/env python
####################################################################################
# Copyright (c) 2009, Centre for Advanced Internet Architectures
# Swinburne University of Technology, Melbourne, Australia
# (CRICOS number 00111D).
#
####################################################################################
#
# Update Regenerator for the Quagga Accelerator - URQA: a tool for replaying 
# recorded BGP updates in MRT format into live BGP sessions. 
# This version of the UR creates a fake clock and synchronises to Quagga 
# to adapt the speed.
# Requires a 4 byte AS number and Multiprotocol capable dpkt version.
# See README-quagga-accelerator for details.
#
####################################################################################
#
# This software was developed by Mattia Rossi <mrossi@swin.edu.au>
#
# This software has been made possible in part by a grant from
# the Cisco University Research Program Fund at
# Community Foundation Silicon Valley
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The names of the authors, the "Centre for Advanced Internet Architecture"
#    and "Swinburne University of Technology" may not be used to endorse
#    or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
###############################################################################

import dpkt
from pybgpdump import BGPDump
import socket
import threading
import struct
from optparse import OptionParser
import re
import mmap
import os
import time
import errno
import stat

# Create the OPEN message. The capabilities are hardcoded.
def mkOpenPkt(asn,srcip, astype):
    bgppkt=dpkt.bgp.BGP()
    openpkt=bgppkt.Open(v=4, identifier=struct.unpack('!L',socket.inet_aton(srcip))[0],holdtime=180)
    if astype:
        if (asn>65535 ):
            openpkt.asn=23456
        else:
            openpkt.asn=asn
    else:
        openpkt.asn=asn
    openpkt.parameters=[]
    param1=openpkt.Parameter()
    cap1=param1.Capability(code=dpkt.bgp.CAP_MULTIPROTOCOL)
    multi=cap1.MultiProtocol(afi=1, res=0, safi=1)
    cap1.data=multi
    cap1.len=len(multi)
    param1.data=cap1
    param1.type=dpkt.bgp.CAPABILITY
    param1.len=len(cap1)
    openpkt.parameters.append(param1)
    if astype:
        param2=openpkt.Parameter()
        cap2=param2.Capability(code=dpkt.bgp.CAP_SUPPORT_AS4)
        cap2.data=struct.pack("!L", asn)
        cap2.len=len(cap2.data)
        param2.data=cap2
        param2.type=dpkt.bgp.CAPABILITY
        param2.len=len(cap2)
        openpkt.parameters.append(param2)
    param3=openpkt.Parameter()
    cap3=param3.Capability(code=dpkt.bgp.CAP_ROUTE_REFRESH_OLD) # Route Refresh old
    cap3.len=0
    param3.data=cap3
    param3.type=dpkt.bgp.CAPABILITY
    param3.len=len(cap3)
    openpkt.parameters.append(param3)
    param4=openpkt.Parameter()
    cap4=param4.Capability(code=dpkt.bgp.CAP_ROUTE_REFRESH) # Route Refresh old
    cap4.len=0
    param4.data=cap4
    param4.type=dpkt.bgp.CAPABILITY
    param4.len=len(cap4)
    openpkt.parameters.append(param4)
    openpkt.param_len=len(openpkt.parameters)
    #capabilities: AS4, MULTIPROTO for 1/1 and route refresh (old/new)
    bgppkt.data=openpkt.pack()
    bgppkt.len=len(bgppkt)
    return bgppkt.pack()   

def main():
    parser = OptionParser()
    parser.add_option("-m", "--mrtfile", dest="mrtfile", default="", 
                    help="mrt input file")
    parser.add_option("-d", "--dst",
                    dest="dstip", help="destination ip address")
    parser.add_option("-a", "--asip",
                    dest="asip", help="file containing ip address to as number mapping")
    parser.add_option("-q", "--quaggafile",
                    dest="quaggafile", default="", help="file containing the commands to start quagga instances")
    
    (options, args) = parser.parse_args()
      
    if(not options.mrtfile or not options.dstip or not options.asip):
        parser.print_help()
        exit(-1)
      
    dstip=options.dstip
      
    # Read the start time
    ft=open(options.mrtfile, "r")
    starttime=struct.unpack("!I", ft.read(4))[0]

    print "Start time: " + str(starttime) + " - " + time.strftime("%a, %d %b %Y %H:%M:%S",time.localtime(starttime))
    ft.close()
        
    # This should give us enough time to startup all needed quagga instances
    faketime=starttime-90
      
    # Create the fake time file
    mfd=os.open("/tmp/faketime",os.O_RDWR|os.O_NONBLOCK|os.O_CREAT,0777)
    os.write(mfd, struct.pack("I", faketime))
    mm=mmap.mmap(mfd, 4, mmap.ACCESS_COPY)
    mm.write(struct.pack("I", faketime))
    mm.seek(0, os.SEEK_SET)

    srciplist={}

    # Start all Quagga instances listed in the Quagga commands file
    qf=open(options.quaggafile, "r")
    syncfile=""
    synclist=[]
    pidlist=[]
    for line in qf:
        # Skip lines that start with a # (hash)
        m=re.search("^\s*#+",line)
        if m:
            continue
        # Read lines - no syntax check is performed - it shouldn't be too hard to write a proper quagga command
        m=re.search(".*bgpd.*",line)
        if m:
            params=line.split(" ")
            for i in xrange(len(params)):
                params[i]=params[i].strip()
            qcom=params[0]
            params[0]=""
            next=0
            for el in params:
                if next==1:
                    break
                if el=="-s":
                    next=1
            # Create the synchronization file - load feedback
            syncfile=el
            sfd=os.open(syncfile,os.O_RDWR|os.O_NONBLOCK|os.O_CREAT,0777)
            os.write(sfd, '\0')
            map=mmap.mmap(sfd,1,mmap.ACCESS_COPY)
            synclist.append(map)
            pid=os.fork()
            if (pid==0):
                os.execvp(qcom,params)
                os._exit(0)
            pidlist.append(pid)		
            #  Advance the fake clock by 1 sec and sleep a sec - this is realtime
            faketime=faketime+1
            time.sleep(1)
            mm.write(struct.pack("I", faketime))
            mm.seek(0, os.SEEK_SET)

    if synclist==[]:
        print "No quagga instances executed!"
        os._exit(0)

    # Prepare the Keepalive packet
    keepalivetimer=30
    keepalive=dpkt.bgp.BGP(type=dpkt.bgp.KEEPALIVE)
    keepalive.len=len(keepalive)
    
    # Wait for 5 seconds before trying to connect to the first Quagga. Within that time it should easily be ready to accept connections
    for i in xrange(5):
        faketime=faketime+1
        time.sleep(1)
        mm.write(struct.pack("I", faketime))
        mm.seek(0, os.SEEK_SET)

    #Start connecting to quagga
    f=open(options.asip, "r")
    for line in f:
        m=re.search("([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}) ([0-9]+) ([01]).*", line)
        if (m):
            srciplist[m.group(1)]=[int(m.group(2)),int(m.group(3))]
            print "connecting to: " + dstip +" from " + m.group(1) + " asn: " + str(srciplist[m.group(1)][0]) + "\n"
            srciplist[m.group(1)].append(socket.socket(socket.AF_INET, socket.SOCK_STREAM))
            con=1
            while(con != 0): #try to connect to quagga until it works
                srciplist[m.group(1)][2].close()
                srciplist[m.group(1)][2] = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                srciplist[m.group(1)][2].bind((m.group(1), 179))
                con=srciplist[m.group(1)][2].connect_ex((dstip, 179))
            print "connected!\n"
            #Send OPEN message
            srciplist[m.group(1)][2].send(mkOpenPkt(srciplist[m.group(1)][0],m.group(1),srciplist[m.group(1)][1]))
            # Allow a second
            faketime=faketime+1
            time.sleep(1)
            mm.write(struct.pack("I", faketime))
            mm.seek(0, os.SEEK_SET)
            # Send Keepalive - now the BGP session is established!
            srciplist[m.group(1)][2].send(keepalive.pack())

    # Finish the remaining time until we get to teh start time - this could also be done in log time
    while faketime < starttime:
        faketime=faketime+1
        keepalivetimer=keepalivetimer-1
        if not keepalivetimer:
            for e in srciplist.keys():
                srciplist[e][2].send(keepalive.pack())
            keepalivetimer=30
        time.sleep(1)
        mm.write(struct.pack("I", faketime))
        mm.seek(0, os.SEEK_SET)

    # Open MRT file using pybgpdump
    dump = BGPDump(options.mrtfile)
    found=0

    # Variables for statistics
    start=time.time()
    totstart=time.time()
    fakestart=faketime
    reltime=0
    # Start processing MRT file
    print "Sending updates..."
    for mrt_h, bgp_h, bgp_m in dump:
        srcipdot = socket.inet_ntoa(struct.pack('>L',bgp_h.src_ip))
        while (faketime < mrt_h.ts):
            reltime=reltime+1
            stop=time.time()
            speedup=1.0/(stop-start)
            # print "Speedup: " + str(speedup)
            start=time.time()
            # Check for feedback from Quagga
            for sm in synclist:
                readsync=sm.read_byte()
                sm.seek(0, os.SEEK_SET)
                while(readsync=='1'):
                    # If realtime is 1 second, proceed. There is no point in waiting more than realtime
                    if time.time()-start>=1.0:
                        sm.write_byte('\0')
                        sm.seek(0, os.SEEK_SET)
                    readsync=sm.read_byte()
                    sm.seek(0, os.SEEK_SET)
            keepalivetimer=keepalivetimer-1
            if(keepalivetimer==0):
                for e in srciplist.keys():
                    srciplist[e][2].send(keepalive.pack())
                keepalivetimer=30
            mm.write(struct.pack("I", faketime))
            mm.seek(0, os.SEEK_SET) 
            faketime = faketime + 1
        #If  faketime equals mrt_h.ts there is no need to send keepalives or to write into faketime mmap - just send the packet
        if srcipdot in srciplist.keys():
            srciplist[srcipdot][2].send(bgp_h.data)
    totend=time.time()
    fakeend=faketime

    #Clean up after finishing - Allow Quaggas to finish processing
    for i in xrange(30):
        faketime=faketime+1
        time.sleep(1)
        mm.write(struct.pack("I", faketime+i))
        mm.seek(0, os.SEEK_SET) 

    # Close sockets and try to kill Quagga BGP daemons
    for e in srciplist.keys():
        srciplist[e][2].close()
    for pid in pidlist:
    	os.kill(pid,15)

    # Print the total speedup
    totspeedup=float(fakeend-fakestart)/(totend-totstart)
    print "Finished! - Speedup: " +str(totspeedup)

main()
